use nom::branch::alt;
use nom::character::complete::line_ending;
use nom::combinator::eof;
use nom::combinator::recognize;
use nom::combinator::verify;
use nom::multi::many1;
use nom::multi::many_till;
use nom::sequence::tuple;

use crate::parser::object::standard_set_object;
use crate::parser::parser_context::ChainBehavior;
use crate::parser::parser_context::ContextElement;
use crate::parser::parser_context::ExitMatcherNode;
use crate::parser::parser_with_context::parser_with_context;
use crate::parser::util::exit_matcher_parser;

use super::element::non_paragraph_element;
use super::error::Res;
use super::lesser_element::Paragraph;
use super::util::blank_line;
use super::util::get_consumed;
use super::util::trailing_whitespace;
use super::Context;

#[tracing::instrument(ret, level = "debug")]
pub fn paragraph<'r, 's>(context: Context<'r, 's>, input: &'s str) -> Res<&'s str, Paragraph<'s>> {
    let parser_context =
        context.with_additional_node(ContextElement::ExitMatcherNode(ExitMatcherNode {
            exit_matcher: ChainBehavior::AndParent(Some(&paragraph_end)),
        }));
    let standard_set_object_matcher = parser_with_context!(standard_set_object)(&parser_context);
    let exit_matcher = parser_with_context!(exit_matcher_parser)(&parser_context);

    let (remaining, (children, _exit_contents)) = verify(
        many_till(standard_set_object_matcher, exit_matcher),
        |(children, _exit_contents)| !children.is_empty(),
    )(input)?;

    let source = get_consumed(input, remaining);

    Ok((remaining, Paragraph { source, children }))
}

#[tracing::instrument(ret, level = "debug")]
fn paragraph_end<'r, 's>(context: Context<'r, 's>, input: &'s str) -> Res<&'s str, &'s str> {
    let non_paragraph_element_matcher = parser_with_context!(non_paragraph_element)(context);
    alt((
        recognize(tuple((line_ending, many1(blank_line)))),
        recognize(tuple((line_ending, non_paragraph_element_matcher))),
        eof,
    ))(input)
}
